/*
 * Created on 20.12.2004
 *
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
package edu.teco.particles.pComponent;

import java.awt.BorderLayout;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.List;

import javax.swing.Action;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import edu.teco.particles.libParticle.P_FILTER;
import edu.teco.particles.libParticle.ParticleFilter;
import edu.teco.particles.libParticle.ParticlePacket;
import edu.teco.particles.libParticle.ParticleSrcId;
import edu.teco.particles.libParticle.ParticleTuple;
import edu.teco.particles.pComponent.io.ParticlePacketListener;
import edu.teco.particles.pComponent.io.ParticlePacketsEvent;
import edu.teco.particles.pComponent.scanner.AddSrcIdAction;
import edu.teco.particles.pComponent.scanner.ContextTreeModel;
import edu.teco.particles.pComponent.scanner.ParticleNodeRenderer;
import edu.teco.particles.pComponent.util.ParticleContext;

/**
 * The Facade of the Scanner component. Functionality: 
 * - scan the Network for available particles. 
 * - scan the Network for available contextes. 
 * - scanning can be active and passive.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */
public class Scanner extends JScrollPane implements TreeSelectionListener,
    ParticlePacketListener
{

  private DefaultMutableTreeNode top    = new DefaultMutableTreeNode("root");

  protected JTree                _tree  = new JTree();

  protected ContextTreeModel     _model = new ContextTreeModel();

  /**
   * call the super Constructor.
   */
  public Scanner()
  {
    _tree.setCellRenderer(new ParticleNodeRenderer());
    _tree.setLayout(new BorderLayout());
    _tree.setModel(_model);
    _tree.getSelectionModel().setSelectionMode(
        TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
    _tree.addTreeSelectionListener(this);

    final JPopupMenu popup = new JPopupMenu();
    final Action srcId = new AddSrcIdAction(_tree);
    popup.add(srcId);

    _tree.addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e)
      {
        if (e.isPopupTrigger())
        {
          popup.show(e.getComponent(), e.getX(), e.getY());
        }
      }

      public void mouseReleased(MouseEvent e)
      {
        if (e.isPopupTrigger())
        {
          popup.show(e.getComponent(), e.getX(), e.getY());
        }
      }
    });

    setViewportView(_tree);
  }

  public void particlePackets(ParticlePacketsEvent e)
  {
    assert (_socket != null);

    List<ParticlePacket> _buffer = _socket.getBuffer();

    ParticleTuple next;
    ParticleSrcId srcid;
    boolean mustBeUpdated = false;

    synchronized (_buffer)
    {

      for (ParticlePacket p : _buffer)
      {
        srcid = p.getSrcId();
        next = p.firstAcl();

        while (next != null)
        {
          if (_model.add(p.getSrcId(), next)) mustBeUpdated = true;
          next = p.nextAcl(next);
        }
      }
    }

    if (mustBeUpdated) _tree.updateUI();
  }

  /**
   * @name RecvSocketThreadProperty
   * @short Property for the SocketThread from which packets are received. @{
   */
  private SocketThread _socket;

  public SocketThread getRecvSocket()
  {
    return _socket;
  }

  public void setRecvSocket(SocketThread socket)
  {
    assert (socket != null);
    assert (socket != _filterSocket);

    _socket = socket;
    socket.addParticlePacketListener(this);
  }

  /* @} */

  /**
   * @name FilterSocketThreadProperty
   * @short Property for the SocketThread on which the filter is set. @{
   */
  private SocketThread _filterSocket;

  public SocketThread getFilterSocket()
  {
    return _socket;
  }

  public void setFilterSocket(SocketThread socket)
  {
    assert (socket != null);
    assert (socket != _socket);

    _filterSocket = socket;

    TreePath path = new TreePath(new Object[] { _model.getRoot() });
    _tree.getSelectionModel().setSelectionPath(path);
  }

  /** @} */

  protected boolean _justCreatedFilter;

  /**
   * reacts on the selection of something in the list.
   * 
   * We walk through the complete selection on every call and build the filters +
   * If a Tuple is selected, we also change the selection to contain the SrcId
   * the tuple is associated with.
   */
  public void valueChanged(TreeSelectionEvent e)
  {
    assert (_filterSocket != null);
    ParticleFilter filter = new ParticleFilter("");
    ContextTreeModel model = (ContextTreeModel) _tree.getModel();
    _justCreatedFilter = true;

    for (TreePath path : _tree.getSelectionPaths())
    {
      Object node = path.getLastPathComponent();

      // walk through the complete selection
      // if element is root -> build "null" filter -> bail out
      // else add srcid to filter
      // add filter to filtersocket

      if (node.equals(model.getRoot()))
      {
        _filterSocket.setFilter(new ParticleFilter(""));
        break;
      }
      else if (node instanceof ParticleSrcId)
      {
        _addToFilter(filter, (ParticleSrcId) node);
      }
      else if (node instanceof ParticleContext)
      {
        ParticleSrcId id = ((ParticleContext) node).getSrcId();
        _addToFilter(filter, id);

        TreePath srcpath = new TreePath(new Object[] { model.getRoot(), id });
        _tree.getSelectionModel().addSelectionPath(srcpath);
      }
      else
      {
        assert (false) : "code for node not implemented";
      }
    }

    _filterSocket.setFilter(filter);
  }

  /**
   * adds a SrcId to the ParticleFilter.
   * 
   * @param filter
   *          the filter to add the srcid to.
   * @param id
   *          the srcid to build the filter on.
   */
  protected void _addToFilter(ParticleFilter filter, ParticleSrcId id)
  {
    int filterMode;

    if (_justCreatedFilter)
    {
      filterMode = P_FILTER.FILTER_TYPE_ID | P_FILTER.FILTER_CONCAT_AND;
      _justCreatedFilter = false;
    }
    else
    {
      filterMode = P_FILTER.FILTER_TYPE_ID | P_FILTER.FILTER_CONCAT_OR;
    }

    filter.add(filterMode, id.toFilter());
  }

  /*
   * ===========================================================================
   * delegate to JTree
   * ========================================================================
   */
  public void addTreeSelectionListener(TreeSelectionListener list)
  {
    _tree.addTreeSelectionListener(list);
  }

  public void removeTreeSelectionListener(TreeSelectionListener list)
  {
    _tree.removeTreeSelectionListener(list);
  }

  /*
   * ===========================================================================
   * delegate to ContextTreeModel
   * ========================================================================
   */
  /**
   * add your component here, to be updated if new particles arrived on the net.
   */
  public void addTreeModelListener(TreeModelListener list)
  {
    _model.addTreeModelListener(list);
  }

  public void removeTreeModelListener(TreeModelListener list)
  {
    _model.removeTreeModelListener(list);
  }
}
