/*
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */

package edu.teco.particles.pComponent.scanner;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;

import javax.swing.event.EventListenerList;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

import edu.teco.particles.libParticle.ParticleSrcId;
import edu.teco.particles.libParticle.ParticleTuple;
import edu.teco.particles.pComponent.util.ParticleContext;

/**
 * This is a Class that holds all available Contextes in a TreeModel.
 */
public class ContextTreeModel implements TreeModel
{
  protected EventListenerList listeners;

  public ContextTreeModel()
  {
    listeners = new EventListenerList();
  }

  public int getIndexOfChild(Object parent, Object child)
  {
    for (int count = getChildCount(parent), i = 0; i < count; i++)
      if (getChild(parent, i).equals(child)) return i;

    return -1;
  }

  /** Call when there is a new root, which may be null, i.e. not existent. */
  protected void fireNewRoot()
  {
    Object[] pairs = listeners.getListenerList();

    Object root = getRoot();

    /*
     * Undocumented. I think it is the only reasonable/possible solution to use
     * use null as path if there is no root. TreeModels without root aren't
     * important anyway, since JTree doesn't support them (yet).
     */
    TreePath path = (root != null) ? new TreePath(root) : null;

    TreeModelEvent e = null;

    for (int i = pairs.length - 2; i >= 0; i -= 2)
    {
      if (pairs[i] == TreeModelListener.class)
      {
        if (e == null) e = new TreeModelEvent(this, path, null, null);

        ((TreeModelListener) pairs[i + 1]).treeStructureChanged(e);
      }
    }
  }

  /**
   * Call when everything but the root has changed. Only may be called when the
   * root is not null. Otherwise there isn't a structure to have changed.
   */
  protected void fireStructureChanged()
  {
    fireTreeStructureChanged(new TreePath(getRoot()));
  }

  /** Call when a node has changed its leaf state. */
  protected void firePathLeafStateChanged(TreePath path)
  {
    fireTreeStructureChanged(path);
  }

  /** Call when the tree structure below the path has completely changed. */
  protected void fireTreeStructureChanged(TreePath parentPath)
  {
    Object[] pairs = listeners.getListenerList();

    TreeModelEvent e = null;

    for (int i = pairs.length - 2; i >= 0; i -= 2)
    {
      if (pairs[i] == TreeModelListener.class)
      {
        if (e == null) e = new TreeModelEvent(this, parentPath, null, null);

        ((TreeModelListener) pairs[i + 1]).treeStructureChanged(e);
      }
    }
  }

  /**
   * Call when the path itself has changed, but no structure changes have
   * occurred.
   */
  protected void firePathChanged(TreePath path)
  {
    Object node = path.getLastPathComponent();
    TreePath parentPath = path.getParentPath();

    if (parentPath == null)
      fireChildrenChanged(path, null, null);
    else
    {
      Object parent = parentPath.getLastPathComponent();

      fireChildChanged(parentPath, getIndexOfChild(parent, node), node);
    }
  }

  protected void fireChildAdded(TreePath parentPath, int index, Object child)
  {
    fireChildrenAdded(parentPath, new int[] { index }, new Object[] { child });
  }

  protected void fireChildChanged(TreePath parentPath, int index, Object child)
  {
    fireChildrenChanged(parentPath, new int[] { index }, new Object[] { child });
  }

  protected void fireChildRemoved(TreePath parentPath, int index, Object child)
  {
    fireChildrenRemoved(parentPath, new int[] { index }, new Object[] { child });
  }

  protected void fireChildrenAdded(TreePath parentPath, int[] indices,
      Object[] children)
  {
    Object[] pairs = listeners.getListenerList();

    TreeModelEvent e = null;

    for (int i = pairs.length - 2; i >= 0; i -= 2)
    {
      if (pairs[i] == TreeModelListener.class)
      {
        if (e == null)
          e = new TreeModelEvent(this, parentPath, indices, children);

        ((TreeModelListener) pairs[i + 1]).treeNodesInserted(e);
      }
    }
  }

  protected void fireChildrenChanged(TreePath parentPath, int[] indices,
      Object[] children)
  {
    Object[] pairs = listeners.getListenerList();

    TreeModelEvent e = null;

    for (int i = pairs.length - 2; i >= 0; i -= 2)
    {
      if (pairs[i] == TreeModelListener.class)
      {
        if (e == null)
          e = new TreeModelEvent(this, parentPath, indices, children);

        ((TreeModelListener) pairs[i + 1]).treeNodesChanged(e);
      }
    }
  }

  protected void fireChildrenRemoved(TreePath parentPath, int[] indices,
      Object[] children)
  {
    Object[] pairs = listeners.getListenerList();

    TreeModelEvent e = null;

    for (int i = pairs.length - 2; i >= 0; i -= 2)
    {
      if (pairs[i] == TreeModelListener.class)
      {
        if (e == null)
          e = new TreeModelEvent(this, parentPath, indices, children);
        ((TreeModelListener) pairs[i + 1]).treeNodesRemoved(e);
      }
    }
  }

  protected Object clone() throws CloneNotSupportedException
  {
    ContextTreeModel clone = (ContextTreeModel) super.clone();

    clone.listeners = new EventListenerList();

    return clone;
  }

  public void addTreeModelListener(TreeModelListener l)
  {
    listeners.add(TreeModelListener.class, l);
  }

  public void removeTreeModelListener(TreeModelListener l)
  {
    listeners.remove(TreeModelListener.class, l);
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.tree.TreeModel#getRoot()
   */
  public Object getRoot()
  {
    return "all contextes";
  }

  /** contains the mapping from srcid to aclList */
  protected LinkedHashMap<ParticleSrcId, List<ParticleContext>> _map = new LinkedHashMap<ParticleSrcId, List<ParticleContext>>();

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.tree.TreeModel#getChild(java.lang.Object, int)
   */
  public Object getChild(Object parent, int index)
  {
    if (parent.equals(getRoot()))
    {
      return _map.keySet().toArray()[index];
    }
    else if (parent instanceof ParticleSrcId)
    {
      ParticleSrcId srcid = (ParticleSrcId) parent;
      List tupleList = _map.get(srcid);
      return tupleList.get(index);
    }
    else
    {
      return null;
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.tree.TreeModel#getChildCount(java.lang.Object)
   */
  public int getChildCount(Object parent)
  {
    if (parent.equals(getRoot()))
    {
      return _map.size();
    }
    else if (parent instanceof ParticleSrcId)
    {
      ParticleSrcId srcid = (ParticleSrcId) parent;
      List tupleList = _map.get(srcid);
      return tupleList.size();
    }
    else
    {
      return 0;
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.tree.TreeModel#isLeaf(java.lang.Object)
   */
  public boolean isLeaf(Object aNode)
  {
    return (aNode instanceof ParticleContext);
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.swing.tree.TreeModel#valueForPathChanged(javax.swing.tree.TreePath,
   *      java.lang.Object)
   */
  public void valueForPathChanged(TreePath arg0, Object arg1)
  {
    assert (false) : "not implemented ";
  }

  public boolean add(ParticleSrcId id, ParticleTuple tuple)
  {
    assert (tuple != null);
    assert (id != null);

    List<ParticleContext> contextList = _map.get(id);
    ParticleContext current = new ParticleContext(id, tuple);
    TreePath path;
    boolean retValue = false;

    if (contextList == null)
    {
      contextList = new LinkedList<ParticleContext>();
      _map.put(id, contextList);
      assert (_map.containsKey(id));

      Object[] arr = { getRoot(), id };
      path = new TreePath(arr);
      fireChildAdded(path, 0, id);
      retValue = true;
    }

    if (!contextList.contains(current))
    {
      contextList.add(0, current);

      Object[] arr = { getRoot(), id, current };
      path = new TreePath(arr);
      fireChildAdded(path, contextList.size() - 1, current);
      retValue = true;
    }

    return retValue;
  }
}
