/*
 * Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */

package edu.teco.particles.pComponent.scope;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.JComponent;

import edu.teco.particles.pComponent.util.ParticleContext;

/**
 * This is a Canvas where charts are drawn upon.
 * 
 * Current Features: - draw labels, xgrid, ygrid - draw more than one trace *g*
 */
public class Chart extends JComponent implements TraceChangeListener
{
  /** the coordinate system on which to draw the traces */
  private Rectangle2D.Double _traceCoordinate;

  /** size of the border */
  private double             _borderFactor = 0.08;

  public Chart()
  {
    super();
    setMinimumSize(new Dimension(200, -400));
  }

  private boolean painting = false;

  /**
   * redraws this Component
   * 
   * one unclipped Graphics2D, to draw the labels on one clipped Graphics2D, to
   * draw the traces and the Grid on This seperation is needed to make handling
   * the border easier. sync to cancel the current painting if the old hasn't
   * finished yet. XXX hm, the sync is not done correctly i think.
   */
  public void paintComponent(Graphics g)
  {
    _recalcTraceCoordinate();
    assert (_traceCoordinate != null);

    Graphics2D unclipped = (Graphics2D) g;
    Graphics2D clipped = (Graphics2D) unclipped.create();

    // XXX ugly :) don't know how to do it better.
    // associately while resizing this component, the transform gets set
    // to something different than the identity tranformation. This result
    // in wrong drawing of this component. So we make sure that we only
    // redraw while the transform is the identity transformation.
    if (!_getScreenTransform().equals(unclipped.getTransform()))
    {
      repaint();
      return;
    }

    _screenTransform = _getScreenTransform();
    _coordTransform = _getCoordTransform();

    unclipped.setTransform(_coordTransform);
    unclipped.setStroke(new BasicStroke(0.01f));

    clipped.setTransform(_coordTransform);
    clipped.setStroke(new BasicStroke(0.01f));
    clipped.setPaint(Color.white);
    clipped.fill(_traceCoordinate);
    clipped.setClip(_traceCoordinate);

    _drawTraces(clipped);
    if (isDrawYGrid())
    {
      _drawYGrid(unclipped);
    }
    if (isDrawXGrid())
    {
      _drawXGrid(unclipped);
    }
    _drawAxis(unclipped);
  }

  /**
   * calculates a tranformation for on screen coord.
   * 
   * @return the affineTransform.
   */
  protected AffineTransform _getScreenTransform()
  {
    if (_screenTransform == null)
    {
      _screenTransform = new AffineTransform();
    }
    return _screenTransform;
  }

  /**
   * calculates the transformation from data and time coords to screen
   * coordinates.
   * 
   * @return the AffineTransform.
   */
  protected AffineTransform _getCoordTransform()
  {
    AffineTransform result = new AffineTransform();

    double xBorderSize = _borderFactor * _traceCoordinate.getWidth();
    double yBorderSize = _borderFactor * _traceCoordinate.getHeight();

    double xSizeOfFrame = getWidth();
    double ySizeOfFrame = getHeight();

    // calculate scale to fit
    double xfactor = xSizeOfFrame
        / (_traceCoordinate.getWidth() + 2 * xBorderSize);
    double yfactor = ySizeOfFrame
        / (_traceCoordinate.getHeight() + 2 * yBorderSize);
    result.scale(xfactor, -yfactor);

    // calculates the bordering
    double xTrans = _traceCoordinate.getMinX() - xBorderSize;
    double yTrans = _traceCoordinate.getHeight() + _traceCoordinate.getY()
        + yBorderSize;
    result.translate(-xTrans, -yTrans);

    return result;
  }

  protected AffineTransform _screenTransform;

  protected AffineTransform _coordTransform;

  /** draws the axis on the Canvas */
  protected void _drawAxis(Graphics2D g2)
  {
    g2.setTransform(_coordTransform);
    g2.setPaint(Color.black);

    double xAxisPos;

    if (_traceCoordinate.getMinY() > 0)
    {
      xAxisPos = _traceCoordinate.getMinY();
    }
    else if (_traceCoordinate.getMaxY() < 0)
    {
      xAxisPos = _traceCoordinate.getMaxY();
    }
    else
    {
      xAxisPos = 0.0D;
    }

    Point2D.Double xAxisLeft = new Point2D.Double(_traceCoordinate.getMinX(),
        xAxisPos);
    Point2D.Double xAxisRight = new Point2D.Double(_traceCoordinate.getMaxX(),
        xAxisPos);
    Point2D.Double yAxisUpper = new Point2D.Double(_traceCoordinate.getMinX(),
        _traceCoordinate.getMinY());
    Point2D.Double yAxisLower = new Point2D.Double(_traceCoordinate.getMinX(),
        _traceCoordinate.getMaxY());

    Line2D.Double xAxis = new Line2D.Double(xAxisLeft, xAxisRight);
    Line2D.Double yAxis = new Line2D.Double(yAxisUpper, yAxisLower);

    g2.draw(xAxis);
    g2.draw(yAxis);
  }

  /**
   * draws the grid for the Y-Axis on the Canvas. a fixed percentile of the
   * current displayed height should be filled with the grid.
   */
  protected void _drawYGrid(Graphics2D g2)
  {
    g2.setTransform(_coordTransform);

    // the distance between two grid lines
    double distance = _traceCoordinate.getHeight() * percent;
    double x0 = _traceCoordinate.getX();
    double x1 = _traceCoordinate.getWidth() + _traceCoordinate.getX();

    double zeroPoint;

    if (_traceCoordinate.getMinY() > 0)
    {
      zeroPoint = _traceCoordinate.getMinY();
    }
    else if (_traceCoordinate.getMaxY() < 0)
    {
      zeroPoint = _traceCoordinate.getMaxY();
    }
    else
    {
      zeroPoint = 0.0D;
    }

    Font font = g2.getFont();
    FontRenderContext frc = new FontRenderContext(_screenTransform, true, true);
    TextLayout tl;

    Point2D.Double label = new Point2D.Double();

    for (double y = zeroPoint; y <= _traceCoordinate.getMaxY(); y += distance)
    {
      g2.setColor(gridColor);
      g2.draw(new Line2D.Double(x0, y, x1, y));
      g2.setColor(Color.black);
      g2.setTransform(_screenTransform);

      _coordTransform.transform(new Point2D.Double(x0, y), label);

      tl = new TextLayout(((Integer) (int) y).toString(), font, frc);

      g2.translate(label.getX() - tl.getBounds().getWidth() - 5, label.getY()
          - tl.getBounds().getCenterY());
      // g2.rotate( - Math.PI / 4.0 );

      tl.draw(g2, 0.0f, 0.0f);

      // g2.rotate( Math.PI / 4.0 );
      g2.setTransform(_coordTransform);
    }

    for (double y = zeroPoint; y >= _traceCoordinate.getMinY(); y -= distance)
    {
      g2.setColor(gridColor);
      g2.draw(new Line2D.Double(x0, y, x1, y));
      g2.setColor(Color.black);
      g2.setTransform(_screenTransform);

      _coordTransform.transform(new Point2D.Double(x0, y), label);

      tl = new TextLayout(((Integer) (int) y).toString(), font, frc);
      g2.translate(label.getX() - tl.getBounds().getWidth() - 5, label.getY()
          - tl.getBounds().getCenterY());
      // g2.rotate( -Math.PI / 4.0 );

      tl.draw(g2, 0, 0);

      // g2.rotate( Math.PI / 4.0 );
      g2.setTransform(_coordTransform);
    }
  }

  /**
   * draws the grid for the X-Axis on the Canvas.
   * 
   * @see _drawYGrid
   */
  protected void _drawXGrid(Graphics2D g2)
  {
    g2.setTransform(_screenTransform);

    // the distance between two grid lines
    double distance = _traceCoordinate.getWidth() * percent;
    double y0 = _traceCoordinate.getY();
    double y1 = _traceCoordinate.getHeight() + _traceCoordinate.getY();
    g2.setColor(gridColor);

    Point2D pt1OnScreen = new Point2D.Double();
    Point2D pt1OnTrace = new Point2D.Double();
    Point2D pt2OnScreen = new Point2D.Double();
    Point2D pt2OnTrace = new Point2D.Double();

    double zeroPoint = _traceCoordinate.getX();
    double maxValue = _traceCoordinate.getX() + _traceCoordinate.getWidth();

    Font font = g2.getFont();
    FontRenderContext frc = new FontRenderContext(_screenTransform, true, true);
    TextLayout tl;

    for (double x = zeroPoint; x <= maxValue; x += distance)
    {
      pt1OnTrace.setLocation(x, y0);
      pt2OnTrace.setLocation(x, y1);

      _coordTransform.transform(pt1OnTrace, pt1OnScreen);
      _coordTransform.transform(pt2OnTrace, pt2OnScreen);

      g2.setColor(gridColor);
      g2.draw(new Line2D.Double(pt1OnScreen, pt2OnScreen));

      g2.setColor(Color.black);
      g2.translate(pt1OnScreen.getX(), pt1OnScreen.getY());
      g2.rotate(Math.PI / 6);

      tl = new TextLayout(((Integer) (int) x).toString(), font, frc);
      tl.draw(g2, 0.0f, 0.0f);

      g2.rotate(-Math.PI / 6);
      g2.translate(-pt1OnScreen.getX(), -pt1OnScreen.getY());
    }
  }

  /** recalcs the coordinate syste to draw on */
  protected void _recalcTraceCoordinate()
  {
    double currentX, currentY, currentHeight, newWidth;

    currentY = _minY;
    currentHeight = _maxY - _minY;

    if (_traceCoordinate != null)
    {
      currentX = _traceCoordinate.getX();
      newWidth = _resolution * getWidth();
    }
    else
    {
      currentX = 0.0;
      newWidth = _resolution * getWidth();
    }

    /** rectangle created with x, y, width, height */
    _traceCoordinate = new Rectangle2D.Double(currentX, currentY, newWidth,
        currentHeight);
  }

  /**
   * draws the traces synchronized because we don't want to add/remove the
   * traces while iterating them.
   */
  protected synchronized void _drawTraces(Graphics2D g2)
  {
    g2.setTransform(_coordTransform);

    for (Trace trace : TraceFactory.getAllTraces())
    {
      if (trace.isVisible()) _drawTrace(trace, g2);
    }
  }

  /** draws a trace */
  protected void _drawTrace(Trace trace, Graphics2D g)
  {
    g.setPaint((Paint) trace.getColor());

    if (isDrawImpulses())
    {
      g.draw(TraceShaper.createImpulsesShape(trace, _traceCoordinate,
          _resolution));
    }

    if (isDrawContinious())
    {
      g.draw(TraceShaper.createContiniousShape(trace, _traceCoordinate,
          _resolution));
    }
  }

  /**
   * @name TraceHandling
   * @brief add/remove traces @{
   */

  /**
   * adds a Context that shall be drawn.
   * 
   * @param ctx
   *          the Context the trace is associated with.
   */
  public synchronized void add(ParticleContext ctx)
  {
    Trace trace = TraceFactory.getTrace(ctx);
    trace.addTraceChangeListener(this);
    trace.setVisible(true);
    repaint();
  }

  /**
   * removes a Context from the Chart.
   * 
   * @param ctx
   *          the context wich shall be removed.
   */
  public synchronized void remove(ParticleContext ctx)
  {
    Trace trace = TraceFactory.getTrace(ctx);
    trace.removeTraceChangeListener(this);
    trace.setVisible(false);
    // TraceFactory.removeTrace( ctx );
    repaint();
  }

  /**
   * gets a Trace for a specific context. Only if the context is avaiable on the
   * Chart.
   * 
   * @param context
   *          the context to get the trace for
   * @return the associated Trace or null;
   */
  public Trace getTrace(ParticleContext context)
  {
    if (TraceFactory.getAllContextes().contains(context))
    {
      return TraceFactory.getTrace(context);
    }
    else
    {
      return null;
    }
  }

  /** @} */

  /**
   * @name ResolutionProperty
   * @brief defines the Resolution Property. The Resolution of the Scope is the
   *        interval in milliseconds between subsequent sample points of the
   * trace. So setting the Reolution to 13ms draws a point every 13ms. @{
   */
  protected double _resolution = 13.0;

  public double getResolution()
  {
    return _resolution;
  }

  public void setResolution(double res)
  {
    _resolution = res;
    _recalcTraceCoordinate();
    repaint();
  }

  /** @} */

  /**
   * @name maxYProperty
   * @brief defines the maximum value on the Y axis @{
   */
  protected int _maxY = 100;

  public int getMaxY()
  {
    return _maxY;
  }

  public void setMaxY(int max)
  {
    if (max <= _minY)
    {
      throw new IllegalArgumentException("max value is smaller than min");
    }
    _maxY = max;
    repaint();
  }

  /** @} */

  /**
   * @name minYProperty
   * @brief defines the minimum value on the X axis @{
   */
  protected int _minY = -100;

  public int getMinY()
  {
    return _minY;
  }

  public void setMinY(int min)
  {
    if (min >= _maxY)
    {
      throw new IllegalArgumentException("min value is bigger than max");
    }
    _minY = min;
    repaint();
  }

  /** @} */

  /** slides the coordinate system to the left */
  public void traceChanged(TraceChangedEvent e)
  {
    // needed to set the new rect
    double currentX = _traceCoordinate.getX();
    double currentY = _traceCoordinate.getY();
    double currentWidth = _traceCoordinate.getWidth();
    double currentHeight = _traceCoordinate.getHeight();

    // needed to calculate the offset
    double currentMaxX = _traceCoordinate.getMaxX();

    if (e.getPoint().getX() >= currentMaxX)
    {
      double offset = e.getPoint().getX() - currentMaxX;
      _traceCoordinate.setRect(currentX + offset, currentY, currentWidth,
          currentHeight);
    }
  }

  /** clears the charts of all data */
  public void clear()
  {
    for (Trace trace : TraceFactory.getAllTraces())
    {
      trace.clear();
    }
    _traceCoordinate = null;
    _recalcTraceCoordinate();
  }

  /**
   * wether or not the trace is in the chart
   * 
   * @param trace
   *          the trace which should be checked.
   * @return true if the chart draws this trace otherwise false.
   */
  public boolean contains(Trace trace)
  {
    return TraceFactory.getAllTraces().contains(trace);
  }

  /**
   * @name drawImpulses
   * @brief determines wether the Chart should be drawn with Impulses or not. @{
   */
  protected boolean _drawImpulses = true;

  public boolean isDrawImpulses()
  {
    return _drawImpulses;
  }

  public void setDrawImpulses(boolean state)
  {
    _drawImpulses = state;
    repaint();
  }

  /** @} */

  /**
   * @name drawContinious
   * @brief determines wether the Chart should be drawn with connected lines. @{
   */
  protected boolean _drawContinious = false;

  public boolean isDrawContinious()
  {
    return _drawContinious;
  }

  public void setDrawContinious(boolean state)
  {
    _drawContinious = state;
    repaint();
  }

  /** @} */

  /**
   * @name drawYGrid
   * @brief defines wether the unit grid on the Y-Axis should be drawn or not. @{
   */
  protected boolean _drawYGrid = true;

  public boolean isDrawYGrid()
  {
    return _drawYGrid;
  }

  public void setDrawYGrid(boolean state)
  {
    _drawYGrid = state;
    if (state || isDrawXGrid())
    {
      _borderFactor = 0.08;
    }
    else
    {
      _borderFactor = 0.01;
    }
    repaint();
  }

  /** @} */

  /**
   * @name drawXGrid
   * @brief defines wether the unit grid on the X-Axis should be drawn or not. @{
   */
  protected boolean _drawXGrid = true;

  public boolean isDrawXGrid()
  {
    return _drawXGrid;
  }

  public void setDrawXGrid(boolean state)
  {
    _drawXGrid = state;
    if (state || isDrawYGrid())
    {
      _borderFactor = 0.08;
    }
    else
    {
      _borderFactor = 0.01;
    }
    repaint();
  }

  /** @} */

  /** the percentile of the window which should be covered by the grid */
  protected double percent   = 0.1D;

  /** the color to draw the unit Grid */
  protected Color  gridColor = new Color(220, 220, 220);

}
