/* Copyright (c) 2004
 * Telecooperation Office (TecO), Universitaet Karlsruhe (TH), Germany.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the Universitaet Karlsruhe (TH) nor the names
 *    of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * @author Philipp Scholl <scholl@teco.edu>
 */

package edu.teco.particles.pComponent.scope;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;

import edu.teco.particles.pComponent.util.RingBuffer;

/**
 * a Trace which is limited to show n points.
 */
public class TraceLtd implements Trace
{
  protected Object              mutex = this;

  /** container for the points */
  protected RingBuffer<Point2D> _container;

  /** the maximum number of points in the trace */
  protected int                 _limit;

  /**
   * creates a new TraceLtd.
   * 
   * @param limit
   *          the number of points to hold in the Buffer.
   */
  public TraceLtd(int limit)
  {
    _container = new RingBuffer<Point2D>(limit);
    _observers = new ArrayList<TraceChangeListener>();

    Random r = new Random();
    Color c = new Color(r.nextInt(255), r.nextInt(255), r.nextInt(255));
    setColor(c);
  }

  public void addPoints(List<Point2D> points)
  {
    _container.addAll(points);
  }

  public void addPoint(Point2D point)
  {
    synchronized (mutex)
    {
      _container.add(point);
      _notifyObservers();
    }
  }

  public void addPoint(double x, double y)
  {
    Point2D pt = new Point2D.Double(x, y);
    this.addPoint(pt);
  }

  /**
   * @name TraceObserver
   * @brief the Observer part of the trace @{
   */

  /** container for all observers */
  private ArrayList<TraceChangeListener> _observers;

  /** notifies observers that the trace has changed */
  private void _notifyObservers()
  {
    Iterator it = _observers.iterator();
    while (it.hasNext())
    {
      TraceChangeListener tr = (TraceChangeListener) it.next();
      TraceChangedEvent even;
      try
      {
        synchronized (mutex)
        {
          even = new TraceChangedEvent(_container.getLast());
        }
        tr.traceChanged(even);
      }
      catch (NoSuchElementException e)
      {
        // no data in the trace
      }
    }
  }

  public void addTraceChangeListener(TraceChangeListener obs)
  {
    _observers.add(obs);
  }

  public void removeTraceChangeListener(TraceChangeListener obs)
  {
    _observers.remove(obs);
  }

  public void clear()
  {
    synchronized (mutex)
    {
      _container.clear();
    }
  }

  /**
   * returns an Iterator which return alls points with a minimum difference of
   * to the point before. Following this policy the first point in the trace is
   * only returned if you set a delta of 0.0.
   * 
   * @param delta
   *          the minimum differnce to the last point.
   * @return DeltaIterator Iterator which only returns the points with a
   *         difference >= delta to the last point.
   */
  public Iterator deltaIterator(double delta)
  {
    return new Itr(delta);
  }

  /** an Implementation of a delta Iterator desribet at deltaIteratoru */
  protected class Itr implements Iterator
  {
    /** Index of element to be returned by subsequent call to next. */
    private int      _cursor = 0;

    /** holds the diff which has to be met */
    private double   _delta  = 0.0;

    private Iterator back;

    public Itr(double delta)
    {
      assert (delta > 0.0);
      _delta = delta;
      back = _container.iterator();
    }

    public boolean hasNext()
    {
      return back.hasNext();
    }

    public Point2D next()
    {
      double diff = 0.0;
      Point2D current = null;

      try
      {
        while ((diff - _delta) <= -0.000001)
        {
          current = ((Point2D) back.next());
          diff += current.getX();
          _cursor++;
        }

        return current;
      }
      catch (IndexOutOfBoundsException e)
      {
        return current;
      }
    }

    public void remove()
    {
    }
  }

  private Color _color;

  public Color getColor()
  {
    return _color;
  }

  public void setColor(Color color)
  {
    _color = color;
    _notifyObservers();
  }

  private int _offset = 0;

  public int getOffset()
  {
    return _offset;
  }

  public void setOffset(int offset)
  {
    _offset = offset;
    _notifyObservers();
  }

  protected boolean _visible = true;

  public void setVisible(boolean state)
  {
    _visible = state;
    _notifyObservers();
  }

  public boolean isVisible()
  {
    return _visible;
  }

  /*
   * ===========================================================================
   * Property Observer part of this class delegate to PropertyChangeSupport.
   * ========================================================================
   */
  protected PropertyChangeSupport _supp = new PropertyChangeSupport(this);

  public void addPropertyChangeListener(PropertyChangeListener list)
  {
    _supp.addPropertyChangeListener(list);
  }

  public void removePropertyChangeListener(PropertyChangeListener list)
  {
    _supp.removePropertyChangeListener(list);
  }
}
